# Generate buildbot master.cfg entries.
# $Id$
# Parses a dependency configuration file followed by a master.cfg prototype annotated
# with ##OUTPUT-WORKERS, ##OUTPUT-FACTORIES, ##OUTPUT-SCHEDULERS, and ##OUTPUT-BUILDERS comments
# in the appropriate locations.
#
# The dependency configuration file looks similar to:
#
#   # format: WORKER: os-arch
#   WORKER: fedora-x86_64
#   WORKER: fedora2-x86_64 Linux.x86_64
#   WORKER: fedora-aarch64
#   WORKER: mingw64
#   WORKER: darwin-x86_64
#   WORKER: openbsd-x86_64
#   WORKER: ubuntu-s390x
#   #WORKER: armv7l
#   
#   # format: BUILD-ORDER: project:buildType:worker dependencies
#   BUILD-ORDER: sandboxTools:update:fedora-x86_64
#   BUILD-ORDER: sandboxTools:release:fedora-x86_64 sandboxTools:update:fedora-x86_64
#   BUILD-ORDER: vista:update:fedora-x86_64
#   BUILD-ORDER: vista:documentation:fedora-x86 vista:update:fedora-x86_64
#   BUILD-ORDER: vista:release:fedora-x86_64 vista:documentation:fedora-x86_64
#   BUILD-ORDER: vista:release:mingw64 vista:release:fedora-x86_64
#   BUILD-ORDER: vista:release:darwin-x86_64 vista:release:fedora-x86_64
#   BUILD-ORDER: vista:release:fedora-aarch64 vista:release:fedora-x86_64
#   BUILD-ORDER: vista:release:openbsd-x86_64 vista:release:fedora-x86_64
#   BUILD-ORDER: vista:release:ubuntu-s390x vista:release:mingw64
#   BUILD-ORDER: vista:package:* vista:release:=
#   
#   # These generate factories
#   # Default command is sbmake factoryName
#   # format: BUILD-STEP: forBuilderName sandboxName [command [arg [...]]]
#   # with forBuilderName formatter as project:buildType workers
#   BUILD-STEP: vista:update:fedora-x86_64 vista_arm sbmake update
#   BUILD-STEP: vista:release:* vista_arm sbmake default
#   BUILD-STEP: vista:documentation:fedora-x86_64 vista_arm
#   BUILD-STEP: vista:package:* vista_arm

BEGIN {
	dependencyGraphFile = "buildbotSchedulers.dot"; # output filename for dot file

	workerSuffix = "-worker"; # makeBuildBotWorker uses "-worker" suffix
	builderSuffix = "_builder";
	factorySuffix = "_factory";
	schedulerSuffix = "_scheduler";
	currentFilename = "NOTSET";
	defaultPassword = "bb-passwd";
	phaseCount = 0;
	workerTotal = 0;
	architectureTotal = 0;
	buildStepTotal = 0;
	buildOrderTotal = 0;

	defaultArchitecture["fedora-x86_64"] = "Linux.x86_64";
	defaultArchitecture["fedora-aarch64"] = "Linux.aarch64";
	defaultArchitecture["ubuntu-s390x"] = "Linux.s390x";
	defaultArchitecture["darwin-x86_64"] = "Darwin.x86_64";
	defaultArchitecture["mingw64"] = "Windows.mingw64";
	defaultArchitecture["openbsd-x86_64"] = "OpenBSD.x86_64";
	defaultArchitecture["armv7l"] = "Linux.armv7l";

	dependents[0] = "";
}

function archForWorker(workerName, _n)
{
	if (workerName in workersByName) {
		_n = workersByName[workerName];
		return (workerArch[_n]);
	}
	return (0);

}

/^WORKER: / {
	# format: WORKER: os-cpuArch [archType]
	workerTotal += 1;
	workerNames[workerTotal] = $2;
	workersByName[$2] = workerTotal;
	if (NF >= 3) { # explicit specification of architecture
		archType = $3;
	} else {
		if ($2 in defaultArchitecture) {
			archType = defaultArchitecture[$2];
		} else {
			archType = $2;
		}
	}	
	workerArch[workerTotal] = archType;
	found = 0;
	for(i=1;i<=architectureTotal;i+=1) {
		if (architectureTypes[i] == archType) {
			found = 1;
			break;
		}
	}
	if (found == 0) { # add new architecture
		architectureTotal += 1;
		architectureTypes[architectureTotal] = archType;
		printf("New architecture type: %s\n", archType) > "/dev/stderr";
	}
	if (archType in workersForArchitecture) {
		printf("Additional worker %s for architecture %s\n", $2, archType) > "/dev/stderr";
		workersForArchitecture[archType] = workersForArchitecture[archType] "," $2;
	} else {
		printf("Initial worker %s for architecture %s\n", $2, archType) > "/dev/stderr";
		workersForArchitecture[archType] = $2;
	}
	printf("New worker: %s %s\n", $2, archType) > "/dev/stderr";
	next;
}

/^BUILD-ORDER: / {
	# name format:   projectName:buildName:workerNameOrArchitecture
print $0 > "/dev/stderr";
	n = split($2, elements, ":");
	buildPrefix = elements[1] ":" elements[2] ":";
	targetCount = split(elements[3], targetList, ",");
	for(k=1;k<=targetCount;k+=1) {
printf("target[%d]=%s\n", k, targetList[k]) > "/dev/stderr";
		if (targetList[k] != "*") { # explicit name
			impliedName = buildPrefix targetList[k];
			if (! (impliedName in inferredBuildStepNames)) {
				inferredBuildStepNames[impliedName] = impliedName;
				printf("Added inferred non-wildcard buildStepName of %s\n", impliedName) > "/dev/stderr";
			}
			doExpand = 0;
		} else { # expand
			doExpand = 1;
		}
		if (NF < 3) {
			buildName = buildPrefix targetList[k];
			buildDependentCount[buildName] = 1
			printf("%s has no dependencies\n", buildName) > "/dev/stderr";
			continue;
		}

		for(j=3;j<=NF;j+=1) {
			archList[1] = targetList[k];
			archCount = 1;
			expandThisRound = doExpand;
			n = split($j, elements, ":");
			prefix = elements[1] ":" elements[2] ":";
			do { # this is a restart point
printf("j=%d prefix=%s archCount=%d expand=%d elem=%s arch=%s\n", j, prefix, archCount, expandThisRound, elements[3], archList[1]) > "/dev/stderr";
				restart = 0;
				for(i=1;i<=archCount;i+=1) {
					if (expandThisRound == 0) {
						if (elements[3] == "=") {
							suffix = archList[i]; # take architecture
						} else {
							suffix = elements[3]; # use what was specified
						}
						buildName = buildPrefix archList[i];
					} else { # expand, then redo across archList
						expandThisRound = 0;
						archCount = 0;
						restart = 1;
						l = length(prefix);
						for(b in inferredBuildStepNames) {
							step = inferredBuildStepNames[b];
printf("b=%s step=%s\n", b, step) > "/dev/stderr";
							if (substr(step, 1, l) == prefix) {
								archCount += 1;
								printf("Adding implied architecture %s\n", step) > "/dev/stderr";
								# TODO split step, take element[3]
								n2 = split(step, elems, ":");
								archList[archCount] = elems[3];
							}
						}	
						break;
					}
					if (restart == 1) { # exit for i<=archCount loop
						break;
					}
					name = prefix suffix;
					if (!(name in inferredBuildStepNames)) {
						printf("ERROR: at %d, dependency %s not previously defined\n", NR, name) > "/dev/stderr";
					}
					if (!(buildName in inferredBuildStepNames)) {
						inferredBuildStepNames[buildName] = buildName;
						printf("Added expanded buildStepName of %s\n", buildName) > "/dev/stderr";
					}
					if (buildName in buildDependentCount) { # dependency already defined
						if (doExpand ==  1) {
							printf("At line %d, explicit BUILD-ORDER for %s was already specified, not generating new entry\n", NR, buildName) > "/dev/stderr";
							continue;
						} else {
							printf("ERROR: at line %d, build dependency already specified for %s: %s\n",
		       						NR, buildName, $0) > "/dev/stderr";
						}
					}
					printf("%s dependsUpon %s\n", buildName, name) > "/dev/stderr";
					buildDependents[buildName, j - 2] = name;
					buildDependentCount[buildName] = NF - 1;
					printf("%s dependentCount %d\n", buildName, NF - 2) > "/dev/stderr";
				}
				if (restart == 1) { # only needed to output debug message
					printf("RESTART at %d\n", j) > "/dev/stderr";
					continue;
				}
			} while (restart != 0);
		}
	}

	next;
}

/^BUILD-STEP: / {
	if (NF <= 2) {
		printf("ERROR: at line %d, too few arguments for BUILD-STEP: %s\n",
		       NR, $0) > "/dev/stderr";
		next;
	}
	builderName = $2;
	n = split(builderName, buildElements, ":");
	project = buildElements[1];
	buildName = buildElements[2];
	prefix = project ":" buildName ":";
	l = length(prefix);
	archCount = split(buildElements[3], archList, ",");
	if (archList[1] == "*") { # unqualified
		archCount = 0;
		for(i in inferredBuildStepNames) {
			name = i;
printf("Look at %s\n", name) > "/dev/stderr";
			if (substr(name, 1, l) == prefix) {
				n2 = split(name, elems, ":");
				archCount += 1;
				archList[archCount] = elems[3];
				printf("expand to %s\n", elems[3]) > "/dev/stderr";
			}
		}
#		for(i=1;i<=architectureTotal;i+=1) {
#			archCount += 1;
#			archList[archCount] = architectureTypes[i]; 
#		}
	}

	sbName = $3;
	if (NF <= 3) {
		cmdName = "sbmake";
		arg = buildName;
	} else {
		cmdName = $4;
		if (NF <= 4) {
			arg = buildName;
		} else {
			arg = $5;
		}
	}
	argList[1] = cmdName;
	argCount = 1;
	if (arg != "NOARG") {
		argCount += 1;
		argList[argCount] = arg;
	}
	for(i=6;i<=NF;i+=1) { # add arg $i
		argCount += 1;
		argList[argCount] = $i;
	}
	for(i=1;i<=archCount;i+=1) {
		fullStepName = prefix archList[i];
		if (fullStepName in knownBuildStepNames) {
			printf("ERROR: at %d, build step %s already declared\n",
			       NR, fullStepName) > "/dev/stderr";
			continue;
		}
		buildStepTotal += 1;
		knownBuildStepNames[fullStepName] = buildStepTotal;
		buildStepNames[buildStepTotal] = fullStepName;
		buildStepArgCount[buildStepTotal] = argCount;
		buildStepSB[buildStepTotal] = sbName;
		if (archList[i] in workersForArchitecture) {
			buildWorkers[buildStepTotal] = workersForArchitecture[archList[i]];
		} else {
			# check to see if worker specified by name
			if (!(archList[i] in workersByName)) {
				printf("WARNING: at %d, no worker known for architecture %s\n", NR, archList[i]) > "/dev/stderr";
			}
			buildWorkers[buildStepTotal] = archList[i];
		}
		for(j=1;j<=argCount;j+=1) {
			buildStepArgs[buildStepTotal,j] = argList[j];
		}
		printf("New BuildStep: %s\n", fullStepName) > "/dev/stderr";
	}
}


/##OUTPUT-WORKERS/ {
	printf("c['workers'] = []\n");
	for(i=1;i<=workerTotal;i+=1) {
		usePW = defaultPassword;
		printf("c['workers'].append(worker.Worker(\"%s%s\", \"%s\"",
		       workerNames[i], workerSuffix, usePW);
		if (length(workerArch[i]) > 0) {
			printf(",\n");
			printf("\tproperties={\"BB_ARCH\":\"%s\"} ", workerArch[i]);
		}
		printf("))\n");
	}
	next;
}

/##OUTPUT-FACTORIES/ {
	for(i=1;i<=buildStepTotal;i+=1) {
		stepName = buildStepNames[i];
		n = split(stepName, elements, ":");
		factoryName = elements[1] "_" elements[2] factorySuffix;
		factoryForStep[stepName] = factoryName;
		if (factoryName in emittedFactories) {
#			printf("Already emitted factory for %s\n", factoryName) > "/dev/stderr";
			continue;
		}
		emittedFactories[factoryName] = stepName;

		printf("%s = util.BuildFactory()\n", factoryName);
		sbName = buildStepSB[i];
		if (sbName == "NOT_SB") {
			notSandbox = 1;
			separator = "";
			printf("%s.addStep(steps.ShellCommand(command=[\n\t",
		       		factoryName);
		} else {
			notSandbox = 0;
			if (substr(sbName, 1, 3) == "bt:") {
				workonCmd = "workon_bt";
				sbArg = "--bt";
				sbName = substr(sbName, 4);
			} else {
				workonCmd = "workon";
				sbArg = "--sb"
			}
			separator = ", ";
			printf("%s.addStep(steps.ShellCommand(command=[\n\t\"%s\", \"%s\", \"%s\",\n",
		       		factoryName, workonCmd, sbArg, sbName);
			printf("\t\"--ta\", util.Interpolate(\"%%(prop:BB_ARCH)s+release\"),\n");
			printf("\t\"-c\"");
		}
		for(j=1;j<=buildStepArgCount[i];j+=1) {
			printf("%s\"%s\"", separator, buildStepArgs[i,j]);
			separator = ", ";
		}
		printf("]))\n");
		printf("\n");
	}
	next;
}

/##OUTPUT-SCHEDULERS/ {
	printf("c['schedulers'] = []\n");
	# Make one pass to emit everything without a dependency
	# These can be emitted in any order
	pendingCount = 0;

	printf("# WARNING: automatically generated by %s\n", "makeBuildbotConfig.awk") > dependencyGraphFile;
	printf("digraph buildbotSchedulers  {\n") >> dependencyGraphFile;
	printf("    node [shape=box];\n") >> dependencyGraphFile;
	printf("    subgraph independentSchedulers { rank = same") >> dependencyGraphFile;
		
	for(i=1;i<=buildStepTotal;i+=1) {
		name = buildStepNames[i];
		# NOTE: count is 1 more than true number of dependents
		count = 1 * buildDependentCount[name]; # make numeric even it not defined
		printf("build step %d dependency count: %s %d\n", i, name, count) > "/dev/stderr";
		if (count > 1) { # has dependencies, handle with Dependent scheduler
			pendingCount += 1;
			pendingList[pendingCount] = name;
			continue;
		}
		n = split(name, elements, ":");
		schedulerName = elements[1] "_" elements[2] "_" elements[3];
		gsub("-", "_", schedulerName);
		if (count == 0) {
			printf("WARNING: no BUILD-ORDER defined for %s\n", name) > "/dev/stderr";
		}
		printf("; \"%s\"", name) >> dependencyGraphFile;
		printf("%s%s = schedulers.SingleBranchScheduler(\n", schedulerName, schedulerSuffix);
		printf("\tname=\"%s\",\n", name);
		printf("\tchange_filter=util.ChangeFilter(project=\"%s\"),\n", elements[1]);
		printf("\ttreeStableTimer=None,\n");
		printf("\tbuilderNames=[");
		printf("\"%s%s\"", name, builderSuffix);
		printf("])\n");
		printf("c['schedulers'].append(%s_scheduler)\n", schedulerName);
		printf("\n");
		emittedSchedulerWithName[name] = schedulerName;
	}
	# close subgraph
	printf (" }\n") >> dependencyGraphFile;
	printf ("# dependent schedulers\n") >> dependencyGraphFile;
	# Now make pass collecting dependencies
	emitOrderCount = 0;
	nextElementIndex = 0;
	for(i=1;i<=pendingCount;i+=1) {
		name = pendingList[i];
		if (name in emittedSchedulerWithName) { # already handled
			printf("SKIP %s as already emitted\n", name) > "/dev/stderr";
			continue;
		}
		# push name
		nextElementIndex += 1;
		nameStack[nextElementIndex] = name;
		while (nextElementIndex > 0) {
printf("nextElementIndex=%d\n", nextElementIndex) > "/dev/stderr";
			curName = nameStack[nextElementIndex];
			if (curName in emittedSchedulerWithName) { # already handled
printf("POP %s as now handled\n", curName) > "/dev/stderr";
				nextElementIndex -= 1; # pop it
				continue;
			}

			# NOTE: count is 1 more than true number of dependents
			count = 1 * buildDependentCount[curName]; # make numeric even it not defined
			printf("PENDING entry %d dependency count: %s %d\n", i, curName, count) > "/dev/stderr";
			allDependenciesDone = 1; # assume true
			for(j=1;j<count;j+=1) {
				dependentName = buildDependents[curName, j];
				if (!(dependentName in emittedSchedulerWithName)) { # push dependent name
					allDependenciesDone = 0;
					printf("PUSH dependency %s\n", dependentName) > "/dev/stderr";
					nextElementIndex += 1;
					nameStack[nextElementIndex] = dependentName;
				}
			}
			if (allDependenciesDone == 0) { # restart processing
				printf("Restart scheduler loop with %s\n", nameStack[nextElementIndex]) > "/dev/stderr";
				continue;
			}
			# all dependencies handled

			n = split(curName, elements, ":");
			schedulerName = elements[1] "_" elements[2] "_" elements[3];
			gsub("-", "_", schedulerName);
			emittedSchedulerWithName[curName] = schedulerName;
			for(j=1;j<count;j+=1) {
				dependentName = buildDependents[curName, j];
				printf("\t%s scheduler dependent on %s\n", curName, dependentName) > "/dev/stderr";
				printf("    \"%s\" -> \"%s\";\n", dependentName, curName) >> dependencyGraphFile;
				n = split(dependentName, elements, ":");
				dependentScheduler = elements[1] "_" elements[2] "_" elements[3];
				gsub("-", "_", dependentScheduler);

				printf("%s%s = schedulers.Dependent(\n", schedulerName, schedulerSuffix);
				printf("\tname=\"%s\",\n", schedulerName);
				printf("\tupstream=%s%s,\n", dependentScheduler, schedulerSuffix);
				printf("\tbuilderNames=[");
				printf("\"%s%s\"", curName, builderSuffix);
				printf("])\n");
				printf("c['schedulers'].append(%s%s)\n\n", schedulerName, schedulerSuffix);
			}
		}
	}
	printf ("}\n") >> dependencyGraphFile;
	next;
}

/##OUTPUT-BUILDERS/ {
	printf("c['builders'] = []\n");
	for(i=1;i<=buildStepTotal;i+=1) {
		name = buildStepNames[i];
		factory = factoryForStep[name];
		n = split(name, elements, ":");
		worker_names = ""
		w = split(buildWorkers[i], worker_list, ",");
		printf("c['builders'].append(util.BuilderConfig(\n");
		printf("\tname=\"%s%s\",\n", name, builderSuffix);
	       	printf("\tworkernames=[");
		for(j=1;j<=w;j+=1) {
			if (j != 1) printf(", ");
			printf("\"%s%s\"", worker_list[j], workerSuffix);
		}
		printf("],\n");
		printf("\tfactory=%s))\n\n", factory);
	}
	next;
}

{
	if (FILENAME != currentFilename) {
		phaseCount += 1;
		currentFilename = FILENAME;
		printf("phase %d filename %s\n", phaseCount, currentFilename) > "/dev/stderr";
	}
	if (phaseCount == 2) {
		# by default, output line
		print $0;
		next;
	}
}
